# WANDB run
# https://wandb.ai/scformer-team/scGPT-public/runs/m8ds6s73?workspace=user-
# %%
import copy
import gc
import json
import os
from pathlib import Path
import sys
import time
import traceback
#from typing import List, Tuple, Dict, Union, Optional
import warnings

import torch
from anndata import AnnData
import scanpy as sc
import scvi
import numpy as np
import wandb
from scipy.sparse import issparse
import matplotlib.pyplot as plt
from torch import nn
from torch.nn import functional as F
from torch.utils.data import Dataset, DataLoader
from sklearn.model_selection import train_test_split
from torchtext.vocab import Vocab
from torchtext._torchtext import (
    Vocab as VocabPybind,
)
import pandas as pd
import scipy.sparse as sp
import episcanpy as epi

sys.path.insert(0, "../")
from scgpt import prepare_data, prepare_dataloader, define_wandb_metrcis, evaluate, eval_testdata, train
from scgpt.tokenizer import tokenize_and_pad_batch
from scgpt.preprocess import Preprocessor

import scgpt as scg
from scgpt.tokenizer.gene_tokenizer import GeneVocab
from scgpt.model import TransformerModel, AdversarialDiscriminator
from scgpt.tokenizer import random_mask_value
from scgpt.loss import (
    masked_mse_loss,
    masked_relative_error,
    criterion_neg_log_bernoulli,
)
#from scgpt.preprocess import Preprocessor
from scgpt.utils import set_seed, category_str2int, eval_scib_metrics

sc.set_figure_params(figsize=(4, 4))
os.environ["KMP_WARNINGS"] = "off"
warnings.filterwarnings('ignore')

# %% [markdown]
# ## Step1: Specify hyper-parameter setup for integration task
# Here we provide some hyper-parameter recommendations here for the integration task. #### TODO Note that the PBMC 10K dataset contains multiple batches to be integrated. Therefore, in addition to the default gene modelling objectives, we also turn on ESC, DAR and DSBN objectives specifically to faciliate batch integration.

# %%
hyperparameter_defaults = dict(
    task = 'multiomic',
    seed=0,
    dataset_name="ASAP_PBMC", # Dataset name
    do_train=True, # Flag to indicate whether to do update model parameters during training
    load_model="/scratch/ssd004/datasets/cellxgene/save/cellxgene_census_human-May23-08-36-2023",
    #"../save/scGPT_human", # Path to pre-trained model
    freeze = False, #freeze
    GEP=True, # Gene expression modelling
    GEPC=True, # Gene expression modelling for cell objective
    CLS=False,
    ESC=False,
    DAR = True, # DAR objective weight for batch correction
    DSBN = False,  # Domain-spec batchnorm,
    mask_ratio=0.4, # Default mask ratio
    explicit_zero_prob = False,  # whether explicit bernoulli for zeros
    ecs_thres=0,  # Elastic cell similarity objective, 0.0 to 1.0, 0.0 to disable
    dab_weight=1.0,
    use_batch_labels = True,
    use_mod = True,
    per_seq_batch_sample = False,
    epochs=25, # Default number of epochs for fine-tuning
    input_layer_key = "X_binned", # Default expression value binning in data pre-processing
    n_bins=51, # Default number of bins for value binning in data pre-processing
    n_hvg = 1200,  # Default number of highly variable genes
    n_hvp = 4000,
    max_seq_len = 4001, # # Default n_hvg+1
    lr=1e-3, # Default learning rate for fine-tuning
    batch_size=16, # Default batch size for fine-tuning
    layer_size=512,
    nlayers=4,
    nhead=8, # if load model, batch_size, layer_size, nlayers, nhead will be ignored
    dropout=0.2, # Default dropout rate during model fine-tuning
    schedule_ratio=0.95,  # Default rate for learning rate decay
    save_eval_interval=5, # Default model evaluation interval
    log_interval=100, # Default log interval
    fast_transformer=True, # Default setting
    pre_norm=False, # Default setting
    amp=True,  # Default setting: Automatic Mixed Precision
    pad_token = "<pad>",
    mask_value = -1,
    pad_value = -2,
    include_zero_gene = False,
    even_binning = False,
)

# %%
run = wandb.init(
    config=hyperparameter_defaults,
    project="scGPT",
    reinit=True,
    settings=wandb.Settings(start_method="fork"),
)
config = wandb.config
print(config)

set_seed(config.seed)

# %%
# settings for input and preprocessing
special_tokens = [config.pad_token, "<cls>", "<eoc>"]

# %%
dataset_name = config.dataset_name
save_dir = Path(f"./save/dev_{dataset_name}-{time.strftime('%b%d-%H-%M')}/")
save_dir.mkdir(parents=True, exist_ok=True)
print(f"save to {save_dir}")
logger = scg.logger
scg.utils.add_file_handler(logger, save_dir / "run.log")

# %% [markdown]
# ## Step 2: Load and pre-process data

# %% [markdown]
# ### 2.1 Load the ASAP PBMC data

# %%
if dataset_name == 'ASAP_PBMC':
    GxC1 = np.array(sp.load_npz('/scratch/ssd004/scratch/chloexq/data/multiomic_data/ASAP-PBMC/GxC1.npz').todense().T)
    PxC1 = np.array(sp.load_npz('/scratch/ssd004/scratch/chloexq/data/multiomic_data/ASAP-PBMC/PxC1.npz').todense().T)
    GxC2 = np.array(sp.load_npz('/scratch/ssd004/scratch/chloexq/data/multiomic_data/ASAP-PBMC/GxC2.npz').todense().T)
    PxC2 = np.array(sp.load_npz('/scratch/ssd004/scratch/chloexq/data/multiomic_data/ASAP-PBMC/PxC2.npz').todense().T)
    PxC3 = np.array(sp.load_npz('/scratch/ssd004/scratch/chloexq/data/multiomic_data/ASAP-PBMC/PxC3.npz').todense().T)
    RxC3 = np.array(sp.load_npz('/scratch/ssd004/scratch/chloexq/data/multiomic_data/ASAP-PBMC/RxC3.npz').todense().T)
    PxC4 = np.array(sp.load_npz('/scratch/ssd004/scratch/chloexq/data/multiomic_data/ASAP-PBMC/PxC4.npz').todense().T)
    RxC4 = np.array(sp.load_npz('/scratch/ssd004/scratch/chloexq/data/multiomic_data/ASAP-PBMC/RxC4.npz').todense().T)
    df_genes = pd.read_csv('/scratch/ssd004/scratch/chloexq/data/multiomic_data/ASAP-PBMC/genes.txt', header=None)
    df_proteins = pd.read_csv('/scratch/ssd004/scratch/chloexq/data/multiomic_data/ASAP-PBMC/proteins.txt', header=None)
    df_regions = pd.read_csv('/scratch/ssd004/scratch/chloexq/data/multiomic_data/ASAP-PBMC/regions.txt', header=None)
    celltypes_c1 = pd.read_csv('/scratch/ssd004/scratch/chloexq/data/multiomic_data/ASAP-PBMC/meta_c1.csv')
    celltypes_c2 = pd.read_csv('/scratch/ssd004/scratch/chloexq/data/multiomic_data/ASAP-PBMC/meta_c2.csv')
    celltypes_c3 = pd.read_csv('/scratch/ssd004/scratch/chloexq/data/multiomic_data/ASAP-PBMC/meta_c3.csv')
    celltypes_c4 = pd.read_csv('/scratch/ssd004/scratch/chloexq/data/multiomic_data/ASAP-PBMC/meta_c4.csv')
 
    celltypes_c12 = celltypes_c1['coarse_cluster'].values.tolist()+celltypes_c2['coarse_cluster'].values.tolist()
    adata = AnnData(
        X=np.concatenate([GxC1, GxC2], axis=0),
        obs=pd.DataFrame(celltypes_c12, columns=['cell_type']),
        var=pd.DataFrame(index=df_genes.loc[:, 0].values),
    )
    celltypes_c34 = celltypes_c3['coarse_cluster'].values.tolist()+celltypes_c4['coarse_cluster'].values.tolist()
    adata_atac = AnnData(
        X=np.concatenate([RxC3, RxC4], axis=0),
        obs=pd.DataFrame(celltypes_c34, columns=['cell_type']),
        var=pd.DataFrame(index=df_regions.loc[:, 0].values),
    )
    
    celltypes_combined = celltypes_c12+celltypes_c34
    protein_names = ['p'+ i for i in df_proteins.loc[:, 0].values.tolist()]
    adata_protein = AnnData(
        X=np.concatenate([PxC1, PxC2, PxC3, PxC4], axis=0),
        obs=pd.DataFrame(celltypes_combined, columns=['cell_type']),
        var=pd.DataFrame(index=protein_names),
    )
    
    adata.obs["celltype"] = adata.obs["cell_type"].astype(str).astype('category')
    adata_atac.obs["celltype"] = adata_atac.obs["cell_type"].astype(str).astype('category')
    adata_protein.obs["celltype"] = adata_protein.obs["cell_type"].astype(str).astype('category')
    
    adata.obs["batch_id"] =  np.array([0]*PxC1.shape[0]+ [1]*PxC2.shape[0])
    adata.obs["str_batch"] = np.array([str(i) for i in adata.obs["batch_id"]])
    adata.obs["str_batch"] = adata.obs["str_batch"].astype('category')
    adata_atac.obs["batch_id"] =  np.array([2]*PxC3.shape[0]+ [3]*PxC4.shape[0])
    adata_atac.obs["str_batch"] = np.array([str(i) for i in adata_atac.obs["batch_id"]])
    adata_atac.obs["str_batch"] = adata_atac.obs["str_batch"].astype('category')

    adata.var["gene_name"] = adata.var.index.tolist()
    data_is_raw = True

# %%
if config.use_mod:
    gene_rna_df = pd.DataFrame(index=df_genes.loc[:, 0].values)
    gene_rna_df['mod'] = 'RNA'
    gene_atac_df = pd.DataFrame(index=df_regions.loc[:, 0].values)
    gene_atac_df['mod'] = 'ATAC'
    gene_protein_df = pd.DataFrame(index=['p' + i for i in df_proteins.loc[:, 0].values])
    gene_protein_df['mod'] = 'Protein'
    gene_loc_df = pd.concat([gene_rna_df, gene_atac_df, gene_protein_df])
    gene_loc_df['mod'] = gene_loc_df['mod'].astype('category')

# %% [markdown]
# ### 2.2 Cross-check gene set with the pre-trained model 
# Note that we retain the common gene set between the data and the pre-trained model for further fine-tuning.

# %%
if config.load_model is not None:
    model_dir = Path(config.load_model)
    model_config_file = model_dir / "args.json"
    model_file = model_dir / "best_model.pt"
    vocab_file = model_dir / "vocab.json"

    vocab = GeneVocab.from_file(vocab_file)
    for s in special_tokens:
        if s not in vocab:
            vocab.append_token(s)

    adata.var["id_in_vocab"] = [
        1 if gene in vocab else -1 for gene in adata.var["gene_name"]
    ]
    gene_ids_in_vocab = np.array(adata.var["id_in_vocab"])
    logger.info(
        f"match {np.sum(gene_ids_in_vocab >= 0)}/{len(gene_ids_in_vocab)} genes "
        f"in vocabulary of size {len(vocab)}."
    )
    if config.task in ['integration', 'annotation']:
        adata = adata[:, adata.var["id_in_vocab"] >= 0]
    
    # model
        with open(model_config_file, "r") as f:
            model_configs = json.load(f)
        logger.info(
            f"Resume model from {model_file}, the model args will be overriden by the "
            f"config {model_config_file}."
        )
        embsize = model_configs["embsize"]
        nhead = model_configs["nheads"]
        d_hid = model_configs["d_hid"]
        nlayers = model_configs["nlayers"]
        n_layers_cls = model_configs["n_layers_cls"]
    else:
        embsize = config.layer_size 
        nhead = config.nhead
        nlayers = config.nlayers  
        d_hid = config.layer_size
        old_vocab = vocab
else:
    embsize = config.layer_size 
    nhead = config.nhead
    nlayers = config.nlayers  
    d_hid = config.layer_size

# %% [markdown]
# ### 2.3 Pre-process the data
# We follow the standardized pipline of depth normalization, log normalization, and highly vairable gene (HVG) selection for data pre-processing. We further introduced value binning to obtain the relative expressions of each HVG.

# %% [markdown]
# #### 2.3.1 Pre-process the RNA data

# %%
preprocessor = Preprocessor(
    use_key="X",  # the key in adata.layers to use as raw data
    filter_gene_by_counts=1,  # step 1
    filter_cell_by_counts=1,  # step 2
    normalize_total=1e4,  # 3. whether to normalize the raw data and to what sum
    result_normed_key="X_normed",  # the key in adata.layers to store the normalized data
    log1p=data_is_raw,  # 4. whether to log1p the normalized data
    result_log1p_key="X_log1p",
    subset_hvg=config.n_hvg,  # 5. whether to subset the raw data to highly variable genes
    hvg_flavor="seurat_v3" if data_is_raw else "cell_ranger",
    binning=config.n_bins,  # 6. whether to bin the raw data and to what number of bins
    result_binned_key="X_binned",  # the key in adata.layers to store the binned data
    even_binning=config.even_binning,
)
preprocessor(adata, batch_key=None)

# %% [markdown]
# #### 2.3.2 Pre-process the ATAC data

# %%
# Highly vairable peak selection
epi.pp.filter_cells(adata_atac, min_features = 1)
epi.pp.filter_features(adata_atac, min_cells = 1)
epi.pp.cal_var(adata_atac)
epi.pp.variability_features(adata_atac, nb_features=config.n_hvp, show=False)
adata_atac.raw = adata_atac
adata_atac = epi.pp.select_var_feature(adata_atac, nb_features=config.n_hvp, show=False, copy=True)

# %%
# preprocessor_atac = Preprocessor(
#     use_key="X",  # the key in adata.layers to use as raw data
#     filter_gene_by_counts=0,  # step 1
#     filter_cell_by_counts=False,  # step 2
#     normalize_total=False,  # 3. whether to normalize the raw data and to what sum
#     result_normed_key="X_normed",  # the key in adata.layers to store the normalized data
#     log1p=False,  # 4. whether to log1p the normalized data
#     result_log1p_key="X_log1p",
#     subset_hvg=False,  # 5. whether to subset the raw data to highly variable genes
#     hvg_flavor="seurat_v3" if data_is_raw else "cell_ranger",
#     binning=config.n_bins,  # 6. whether to bin the raw data and to what number of bins
#     result_binned_key="X_binned",  # the key in adata.layers to store the binned data
# )
# preprocessor_atac(adata_atac, batch_key=None)

# %%
def digitize(x: np.ndarray, bins: np.ndarray) -> np.ndarray:
    assert x.ndim == 1 and bins.ndim == 1
    left_digits = np.digitize(x, bins)
    right_difits = np.digitize(x, bins, right=True)
    rands = np.random.rand(len(x))  # uniform random numbers
    digits = rands * (right_difits - left_digits) + left_digits
    digits = np.ceil(digits).astype(np.int64)
    return digits

def do_binning(layer_data):
    binned_rows = []
    bin_edges = []
    for row in layer_data:
        non_zero_ids = row.nonzero()
        non_zero_row = row[non_zero_ids]
        bins = np.quantile(non_zero_row, np.linspace(0, 1, config.n_bins - 1))
        # NOTE: comment this line for now, since this will make the each category
        # has different relative meaning across datasets
        if config.even_binning:
            non_zero_digits = digitize(non_zero_row, bins)
        else:
            non_zero_digits = np.digitize(non_zero_row, bins)
        assert non_zero_digits.min() >= 1
        assert non_zero_digits.max() <= config.n_bins - 1
        binned_row = np.zeros_like(row, dtype=np.int64)
        binned_row[non_zero_ids] = non_zero_digits
        binned_rows.append(binned_row)
        bin_edges.append(np.concatenate([[0], bins]))
    return np.stack(binned_rows)

# %%
layer_data = np.array(adata_atac.X)
adata_atac.obsm["atac_expression_binned"] = do_binning(layer_data)

# %% [markdown]
# #### 2.3.3 Pre-process the Protein data

# %%
# preprocessor_protein = Preprocessor(
#     use_key="X",  # the key in adata.layers to use as raw data
#     filter_gene_by_counts=0,  # step 1
#     filter_cell_by_counts=False,  # step 2
#     normalize_total=False,  # 3. whether to normalize the raw data and to what sum
#     result_normed_key="X_normed",  # the key in adata.layers to store the normalized data
#     log1p=False,  # 4. whether to log1p the normalized data
#     result_log1p_key="X_log1p",
#     subset_hvg=False,  # 5. whether to subset the raw data to highly variable genes
#     hvg_flavor="seurat_v3" if data_is_raw else "cell_ranger",
#     binning=config.n_bins,  # 6. whether to bin the raw data and to what number of bins
#     result_binned_key="X_binned",  # the key in adata.layers to store the binned data
# )
# preprocessor_protein(adata_protein, batch_key=None)

# %%
PxC1_normed = do_binning(PxC1)
PxC2_normed = do_binning(PxC2)
PxC3_normed = do_binning(PxC3)
PxC4_normed = do_binning(PxC4)

# %% [markdown]
# #### 2.3.4 Combine RNA, ATAC, and Protein data

# %%
# n1, d_rna = adata.layers["X_binned"].shape
# _, d_protein = PxC1.shape
# n2, d_atac = adata_atac.layers["X_binned"].shape
# data_RNA_ATAC = np.block(
#         [[adata.layers["X_binned"], -3*np.ones((n1, d_atac))], 
#         [-3*np.ones((n2, d_rna)), adata_atac.layers["X_binned"]]])
# data_combined = np.block([data_RNA_ATAC, adata_protein.layers["X_binned"]])
# var_combined = adata.var.index.tolist() + adata_atac.var.index.tolist() + adata_protein.var.index.tolist()

# adata = AnnData(
#     X=data_combined,
#     obs=pd.concat([adata.obs, adata_atac.obs], axis=0, ignore_index=True),
#     var=pd.DataFrame(index=var_combined),
#     layers={
#         "X_binned": data_combined,}
# )
# adata.var["gene_name"] = adata.var.index.tolist()

# %%
n1, d_rna = adata.layers["X_binned"].shape
_, d_protein = PxC1.shape
n2, d_atac = adata_atac.obsm["atac_expression_binned"].shape
protein_names = ['p'+ i for i in df_proteins.loc[:, 0].values.tolist()]
adata = AnnData(
    X=np.block(
        [[adata.layers["X_binned"], -3*np.ones((n1, d_atac)), np.concatenate([PxC1_normed, PxC2_normed], axis=0)], 
        [-3*np.ones((n2, d_rna)), adata_atac.obsm["atac_expression_binned"], np.concatenate([PxC3_normed, PxC4_normed], axis=0)]]),
    obs=pd.concat([adata.obs, adata_atac.obs], axis=0, ignore_index=True),
    var=pd.DataFrame(index=adata.var.index.tolist() + adata_atac.var.index.tolist() + adata_protein.var.index.tolist()),
    layers={
        "X_binned": np.block(
        [[adata.layers["X_binned"], -3*np.ones((n1, d_atac)), np.concatenate([PxC1_normed, PxC2_normed], axis=0)], 
        [-3*np.ones((n2, d_rna)), adata_atac.obsm["atac_expression_binned"], np.concatenate([PxC3_normed, PxC4_normed], axis=0)]]),}
)

# %%
adata.var["gene_name"] = adata.var.index.tolist()

# %%
adata

# %%
if config.per_seq_batch_sample:
    # sort the adata by batch_id in advance
    adata_sorted = adata[adata.obs["batch_id"].argsort()].copy()

# %% [markdown]
# ### 2.4 Tokenize the input data for model fine-tuning

# %%
all_counts = (
    adata.layers[config.input_layer_key].A
    if issparse(adata.layers[config.input_layer_key])
    else adata.layers[config.input_layer_key]
)
genes = adata.var["gene_name"].tolist()

celltypes_labels = adata.obs["celltype"].tolist()  # make sure count from 0
num_types = len(set(celltypes_labels))
celltypes_labels = np.array(celltypes_labels)

batch_ids = adata.obs["batch_id"].tolist()
num_batch_types = len(set(batch_ids))
batch_ids = np.array(batch_ids)
if config.use_mod:
    mod_type = np.array([gene_loc_df.loc[g, 'mod'] for g in genes])
    vocab_mod = Vocab(
    VocabPybind(np.unique(gene_loc_df['mod']).tolist() + special_tokens, None)
    )
    vocab_mod.set_default_index(vocab_mod["<pad>"])
    mod_type = np.array(vocab_mod(list(mod_type)), dtype=int)
    ntokens_mod = len(vocab_mod)
    

(
    train_data,
    valid_data,
    train_celltype_labels,
    valid_celltype_labels,
    train_batch_labels,
    valid_batch_labels,
) = train_test_split(
    all_counts, celltypes_labels, batch_ids, test_size=0.1, shuffle=True
)


# %%
num_of_non_zero_genes = [
    np.count_nonzero(train_data[i]) for i in range(train_data.shape[0])
]
print(f"max num of non_zero genes: {np.max(num_of_non_zero_genes)}")
print(f"min num of non_zero genes: {np.min(num_of_non_zero_genes)}")
print(f"average num of non_zero genes: {np.mean(num_of_non_zero_genes)}")
print(
    f"99% quantile num of non_zero genes: {np.quantile(num_of_non_zero_genes, 0.99)}"
)
print(f"max original values: {np.max(train_data)}")
print(
    f"average original non_zero values: {np.mean(train_data[np.nonzero(train_data)])}"
)
print(
    f"99% quantile original non_zero values: {np.quantile(train_data[np.nonzero(train_data)], 0.99)}"
)
print(f"num of celltypes: {num_types}")
# TODO
#if zero_gene_tokens > 0:
#max_seq_len = min(max_seq_len, np.max(num_of_non_zero_genes) + 1)
# max_seq_len = min(max_seq_len, len(genes) + 1)
# eval_len = len(genes) + 1
# print(f"set max_seq_len to {max_seq_len}")
# print(f"set eval_len to {eval_len}")


# %%
if config.load_model is None:
    vocab = Vocab(VocabPybind(genes + special_tokens, None))
    vocab.set_default_index(vocab["<pad>"])
    gene_ids = np.array(vocab(genes), dtype=int)
else:
    pretrained_genes = [g for g in genes + special_tokens if g in old_vocab]
    new_genes = [g for g in genes + special_tokens if g not in old_vocab]
    gene_ids_pretrained = np.array(old_vocab(pretrained_genes), dtype=int)
    # https://discuss.pytorch.org/t/expand-an-existing-embedding-and-linear-layer-nan-loss-value/55670/2
    # Retrieve pretrained weights
    vocab = Vocab(VocabPybind(pretrained_genes + new_genes, None))
    vocab.set_default_index(vocab["<pad>"])
    gene_ids = np.array(vocab(genes), dtype=int)

# %%
tokenized_train = tokenize_and_pad_batch(
    train_data,
    gene_ids,
    max_len=config.max_seq_len,
    vocab=vocab,
    pad_token=config.pad_token,
    pad_value=config.pad_value,
    append_cls=True,  # append <cls> token at the beginning
    include_zero_gene=config.include_zero_gene,
    mod_type=mod_type,
    vocab_mod=vocab_mod,
)
tokenized_valid = tokenize_and_pad_batch(
    valid_data,
    gene_ids,
    max_len=config.max_seq_len,
    vocab=vocab,
    pad_token=config.pad_token,
    pad_value=config.pad_value,
    append_cls=True,
    include_zero_gene=config.include_zero_gene,
    mod_type=mod_type,
    vocab_mod=vocab_mod,
)
logger.info(
    f"train set number of samples: {tokenized_train['genes'].shape[0]}, "
    f"\n\t feature length: {tokenized_train['genes'].shape[1]}"
)
logger.info(
    f"valid set number of samples: {tokenized_valid['genes'].shape[0]}, "
    f"\n\t feature length: {tokenized_valid['genes'].shape[1]}"
)

# %%
len(train_data)

# %%
curr_max = 0
for i in range(len(train_data)):
    if np.logical_and(train_data[i, :] != -3, train_data[i, :] != 0).sum() > curr_max:
        curr_max = np.logical_and(train_data[i, :] != -3, train_data[i, :] != 0).sum()

# %%
curr_max

# %%
(adata.X[0, :]==-3).sum()

# %%
tokenized_train['genes'][0]

# %%
(tokenized_train['values']==-2)

# %% [markdown]
# ## Step 3: Load the pre-trained scGPT model
# Note that for multiomic integration, since the pre-trained model does not include the ATAC and protein tokens, we expand the embedding layer by adding these new tokens. We inherit only the gene embedding layer from the pre-trained model, and train rest of the model from scratch.

# %%
# model = TransformerModel(
#     len(old_vocab) if config.load_model is not None else len(vocab),
#     embsize,
#     nhead,
#     d_hid,
#     nlayers,
#     vocab=old_vocab if config.load_model is not None else vocab,
#     dropout=config.dropout,
#     pad_token=config.pad_token,
#     pad_value=config.pad_value,
#     do_mvc=config.GEPC,
#     do_dab=config.DAR,
#     use_batch_labels=config.use_batch_labels,
#     num_batch_labels=num_batch_types,
#     domain_spec_batchnorm=config.DSBN,
#     n_input_bins=config.n_bins,
#     ecs_threshold=config.ecs_thres,
#     explicit_zero_prob=config.explicit_zero_prob,
#     use_fast_transformer=config.fast_transformer,
#     pre_norm=config.pre_norm,
#     use_mod=config.use_mod,
#     ntokens_mod=ntokens_mod if config.use_mod else None,
#     vocab_mod=vocab_mod if config.use_mod else None,
# )

# if config.load_model is not None:
#     try:
#         model.load_state_dict(torch.load(model_file))
#         logger.info(f"Loading all model params from {model_file}")
#     except:
#         # only load params that are in the model and match the size
#         model_dict = model.state_dict()
#         pretrained_dict = torch.load(model_file)
#         pretrained_dict = {
#             k: v
#             for k, v in pretrained_dict.items()
#             if k in model_dict and v.shape == model_dict[k].shape
#         }
#         for k, v in pretrained_dict.items():
#             logger.info(f"Loading params {k} with shape {v.shape}")
#         model_dict.update(pretrained_dict)
#         model.load_state_dict(model_dict)
#     with torch.no_grad():
#         pretrained_emb_weights = model.encoder.embedding.weight.data[gene_ids_pretrained, :].detach().clone()
#         model.encoder.embedding = nn.Embedding(
#             len(vocab), pretrained_emb_weights.shape[1], padding_idx=vocab["<pad>"]
#         )
#         model.encoder.embedding.weight.data[:len(pretrained_genes), :] = pretrained_emb_weights
#     model.vocab = vocab
#     model.ntokens = len(vocab)

# model_dict = torch.load(model_file)

# %%
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
model_dict = torch.load(model_file)
ntokens = len(vocab)  # size of vocabulary
model = TransformerModel(
    ntokens,
    embsize,
    nhead,
    d_hid, 
    nlayers,
    vocab=vocab,
    dropout=config.dropout,
    pad_token=config.pad_token,
    pad_value=config.pad_value,
    do_mvc=config.GEPC,
    do_dab=config.DAR,
    use_batch_labels=config.use_batch_labels,
    num_batch_labels=num_batch_types,
    domain_spec_batchnorm=config.DSBN,
    n_input_bins=config.n_bins,
    ecs_threshold=config.ecs_thres,
    explicit_zero_prob=config.explicit_zero_prob,
    use_fast_transformer=config.fast_transformer,
    pre_norm=config.pre_norm,
    use_mod=config.use_mod,
    ntokens_mod=ntokens_mod if config.use_mod else None,
    vocab_mod=vocab_mod if config.use_mod else None,
)

with torch.no_grad():
    pretrained_emb_weights = model_dict['encoder.embedding.weight'][gene_ids_pretrained, :]
    model.encoder.embedding.weight.data[:len(pretrained_genes), :] = pretrained_emb_weights
    model.encoder.enc_norm.weight.data = model_dict['encoder.enc_norm.weight']
ntokens = len(vocab)

model.to(device)
wandb.watch(model)
print(model)

# %%
if config.GEP and config.GEPC:
    criterion_gep_gepc = masked_mse_loss
if config.CLS:
    criterion_cls = nn.CrossEntropyLoss()
if config.DAR:
    criterion_dab = nn.CrossEntropyLoss()
optimizer = torch.optim.Adam(
    model.parameters(), lr=config.lr, eps=1e-4 if config.amp else 1e-8
)
scheduler = torch.optim.lr_scheduler.StepLR(optimizer, 1, gamma=config.schedule_ratio)
scaler = torch.cuda.amp.GradScaler(enabled=config.amp)

# %% [markdown]
# ## Step 4: Finetune scGPT with task-specific objectives

# %%
best_val_loss = float("inf")
best_avg_bio = 0.0
best_model = None
define_wandb_metrcis()

for epoch in range(1, config.epochs + 1):
    epoch_start_time = time.time()
    train_data_pt, valid_data_pt = prepare_data(
        tokenized_train=tokenized_train, 
        tokenized_valid=tokenized_valid, 
        train_batch_labels=train_batch_labels,
        valid_batch_labels=valid_batch_labels,
        config=config,
        epoch=epoch,
        sort_seq_batch=config.per_seq_batch_sample)
    
    train_loader = prepare_dataloader(
        train_data_pt,
        batch_size=config.batch_size,
        shuffle=True,
        intra_domain_shuffle=False,
        drop_last=False,
        per_seq_batch_sample=config.per_seq_batch_sample
    )
    valid_loader = prepare_dataloader(
        valid_data_pt,
        batch_size=config.batch_size,
        shuffle=False,
        intra_domain_shuffle=False,
        drop_last=False,
        per_seq_batch_sample=config.per_seq_batch_sample
    )

    if config.do_train:
        train(
            model=model,
            loader=train_loader,
            vocab=vocab,
            criterion_gep_gepc=criterion_gep_gepc if config.GEP and config.GEPC else None,
            criterion_dab=criterion_dab if config.DAR else None,
            criterion_cls=criterion_cls if config.CLS else None,
            scaler=scaler,
            optimizer=optimizer,
            scheduler=scheduler,
            device=device,
            config=config,
            logger=logger,
            epoch=epoch,
        )
    val_loss = evaluate(
        model=model,
        loader=valid_loader,
        vocab=vocab,
        criterion_gep_gepc=criterion_gep_gepc if config.GEP and config.GEPC else None,
        criterion_dab=criterion_dab if config.DAR else None,
        criterion_cls=criterion_cls if config.CLS else None,
        device=device,
        config=config,
        epoch=epoch
    )
    elapsed = time.time() - epoch_start_time
    logger.info("-" * 89)
    logger.info(
        f"| end of epoch {epoch:3d} | time: {elapsed:5.2f}s | "
        f"valid loss {val_loss:5.4f} | "
    )
    logger.info("-" * 89)

    if val_loss < best_val_loss:
        best_val_loss = val_loss
        best_model = copy.deepcopy(model)
        best_model_epoch = epoch
        logger.info(f"Best model with score {best_val_loss:5.4f}")

    if epoch % config.save_eval_interval == 0 or epoch == config.epochs:
        logger.info(f"Saving model to {save_dir}")
        torch.save(best_model.state_dict(), save_dir / f"model_e{best_model_epoch}.pt")

        # eval on testdata
        results = eval_testdata(
            model = best_model,
            adata_t = adata_sorted if config.per_seq_batch_sample else adata,
            gene_ids = gene_ids,
            vocab = vocab,
            config = config,
            logger = logger,
            include_types=["cls"],
        )
        results["batch_umap"].savefig(
            save_dir / f"embeddings_batch_umap[cls]_e{best_model_epoch}.png", dpi=300
        )

        results["celltype_umap"].savefig(
            save_dir / f"embeddings_celltype_umap[cls]_e{best_model_epoch}.png", dpi=300
        )
        metrics_to_log = {"test/" + k: v for k, v in results.items()}
        metrics_to_log["test/batch_umap"] = wandb.Image(
            str(save_dir / f"embeddings_batch_umap[cls]_e{best_model_epoch}.png"),
            caption=f"celltype avg_bio epoch {best_model_epoch}",
        )

        metrics_to_log["test/celltype_umap"] = wandb.Image(
            str(save_dir / f"embeddings_celltype_umap[cls]_e{best_model_epoch}.png"),
            caption=f"celltype avg_bio epoch {best_model_epoch}",
        )
        metrics_to_log["test/best_model_epoch"] = best_model_epoch
        wandb.log(metrics_to_log)
        wandb.log({"avg_bio": results.get("avg_bio", 0.0)})

    scheduler.step()

# %%
# save the best model
torch.save(best_model.state_dict(), save_dir / "best_model.pt")

# %%
artifact = wandb.Artifact(f"best_model", type="model")
glob_str = os.path.join(save_dir, "best_model.pt")
artifact.add_file(glob_str)
run.log_artifact(artifact)

run.finish()
wandb.finish()
gc.collect()

# %%



